#pragma once

#include "irc_command.hpp"
#include "ircmsg.hpp"
#include "ratelimit.hpp"
#include "ref.hpp"
#include "stream.hpp"

#include <boost/asio.hpp>
#include <boost/signals2.hpp>

#include <list>
#include <memory>
#include <string>

namespace myirc {

struct SnoteMatch;

class Connection : public std::enable_shared_from_this<Connection>
{
public:
    struct Settings
    {
        bool tls;
        std::string host;
        std::uint16_t port;

        Ref<X509> client_cert;
        Ref<EVP_PKEY> client_key;
        std::string verify;
        std::string sni;

        std::string socks_host;
        std::uint16_t socks_port;
        std::string socks_user;
        std::string socks_pass;
    };

private:
    Stream stream_;
    boost::asio::steady_timer watchdog_timer_;
    std::list<std::string> write_strings_;
    bool write_posted_;

    // Set true when watchdog triggers.
    // Set false when message received.
    bool stalled_;

    /// write buffers after consulting with rate limit
    auto write_buffers() -> void;

    /// write a specific number of messages now
    auto write_buffers(size_t) -> void;

    auto dispatch_line(char *line, bool) -> void;

    static constexpr std::chrono::seconds watchdog_duration = std::chrono::seconds{30};
    auto watchdog() -> void;
    auto watchdog_activity() -> void;

    auto connect(Settings settings) -> boost::asio::awaitable<void>;

public:
    boost::signals2::signal<void(
        boost::asio::ip::tcp::endpoint,
        std::optional<boost::asio::ip::tcp::endpoint>,
        std::string
    )> sig_connect;

    boost::signals2::signal<void(std::exception_ptr)> sig_disconnect;
    boost::signals2::signal<void(IrcCommand, const IrcMsg &, bool flush)> sig_ircmsg;
    std::unique_ptr<RateLimit> rate_limit;

    Connection(boost::asio::io_context &io);

    auto get_executor() -> boost::asio::any_io_executor
    {
        return stream_.get_executor();
    }

    auto start(Settings) -> void;
    auto close() -> void;

    /// Write bytes into the socket.
    auto write_line(std::string message) -> void;

    /// Build and send well-formed IRC message from individual parameters
    auto write_irc(std::string) -> void;
    auto write_irc(std::string, std::string_view) -> void;
    template <typename... Args>
    auto write_irc(std::string front, std::string_view next, Args... rest) -> void;
};

template <typename... Args>
auto Connection::write_irc(std::string front, std::string_view next, Args... rest) -> void
{
    using namespace std::literals;

    if (next.empty() || next.front() == ':' || next.find_first_of("\r\n \0"sv) != next.npos)
    {
        throw std::runtime_error{"bad irc argument"};
    }

    front += " ";
    front += next;
    write_irc(std::move(front), rest...);
}

} // namespace myirc
